/*
******************************************************************************

   LibOVL: (de)alloction functions

   Copyright (C) 2001 Christoph Egger   [Christoph_Egger@t-online.de]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <ggi/internal/ovl.h>
#include <ggi/internal/ovl_debug.h>
#include <ggi/internal/galloc_ops.h>


int ggiOvlAdd(ggiGA_resource_list *reqlist,
	      struct ggiGA_resource_props *prop_motor,
	      struct ggiGA_resource_props *prop_carb,
	      struct ggiGA_resource_props *prop_tank,
	      struct ggiGA_resource_props *prop_tank_cap,
	      enum ggiGA_resource_type type,
	      ggiGA_resource_handle *handle_motor,
	      ggiGA_resource_handle *handle_carb,
	      ggiGA_resource_handle *handle_tank,
	      ggiGA_resource_handle *handle_tank_cap)
{
	int rc = GALLOC_OK;

	if (reqlist == NULL) return GGI_EARGINVAL;
	if (prop_motor == NULL) return GGI_EARGINVAL;
	if (prop_tank == NULL) return GGI_EARGINVAL;
	if (prop_carb == NULL) return GGI_EARGINVAL;
	if (handle_motor == NULL) return GGI_EARGINVAL;
	if (handle_tank == NULL) return GGI_EARGINVAL;
	if (handle_carb == NULL) return GGI_EARGINVAL;


	if (  (!(type & GA_RT_TYPE_MASK) == GA_RT_SPRITE)
	   || (!(type & GA_RT_TYPE_MASK) == GA_RT_VIDEO)
	   || (!(type & GA_RT_TYPE_MASK) == GA_RT_WINDOW))
	{
		return GALLOC_ESPECIFY;
	}	/* if */

	rc = ggiGAAdd(reqlist, prop_motor, (GA_RT_MOTOR | type), handle_motor);
	LIB_ASSERT(rc == GALLOC_OK);

	rc = ggiGAAdd(reqlist, prop_tank, type, handle_tank);
	LIB_ASSERT(rc == GALLOC_OK);

	if (prop_tank_cap != NULL) {
		rc = ggiGAAdd(reqlist, prop_tank_cap, type,
				handle_tank_cap);
		LIB_ASSERT(rc == GALLOC_OK);

		ggiGAAddStateFlag(handle_tank_cap[0], GA_STATE_CAP);
	} else {
		if (handle_tank_cap != NULL) {
			handle_tank_cap[0] = NULL;
		}	/* if */
	}	/* if */

	rc = ggiGAAdd(reqlist, prop_tank, (GA_RT_CARB | type), handle_carb);
	ggiGAAddStateFlag(handle_carb[0], GA_STATE_SEEABOVE);
	LIB_ASSERT(rc == GALLOC_OK);

	rc = ggiGAAttach2MotorGroup(reqlist[0], handle_motor[0], 
				    handle_carb[0]);
	LIB_ASSERT(rc == GALLOC_OK);

	return rc;
}	/* ggiOvlAdd */



int ggiOvlAdd2Group(ggiGA_resource_list *reqlist,
		    struct ggiGA_resource_props *prop_tank,
		    struct ggiGA_resource_props *prop_tank_cap,
		    ggiGA_resource_handle handle_carb,
		    ggiGA_resource_handle *handle_tank,
		    ggiGA_resource_handle *handle_tank_cap)
{
	int rc = GALLOC_OK;
	enum ggiGA_resource_type type;

	if (reqlist == NULL) return GGI_EARGINVAL;
	if (prop_tank == NULL) return GGI_EARGINVAL;
	if (handle_tank == NULL) return GGI_EARGINVAL;
	if (handle_carb == NULL) return GGI_EARGINVAL;

	type = ggiGAGetType(handle_carb);
	type &= GA_RT_TYPE_MASK | GA_RT_SUBTYPE_MASK;
	switch (type &= GA_RT_TYPE_MASK) {
	case GA_RT_SPRITE:
	case GA_RT_VIDEO:
	case GA_RT_WINDOW:
		break;
	default:
		return GALLOC_ESPECIFY;
	}	/* switch */

	rc = ggiGAAdd(reqlist, prop_tank, type, handle_tank);
	LIB_ASSERT(rc == GALLOC_OK);

	if (prop_tank_cap != NULL) {
		rc = ggiGAAdd(reqlist, prop_tank_cap, type,
				handle_tank_cap);
		LIB_ASSERT(rc == GALLOC_OK);

		ggiGAAddStateFlag(handle_tank_cap[0], GA_STATE_CAP);
	} else {
		if (handle_tank_cap != NULL) {
			handle_tank_cap[0] = NULL;
		}	/* if */
	}	/* if */

	rc = ggiGAAppend(handle_carb, reqlist);	
	LIB_ASSERT(rc == GALLOC_OK);

	return rc;
}	/* ggiOvlAdd2Group */



ggiOvl_t ggiOvlCreate(ggi_visual_t vis,
		      struct ggiGA_resource_props *prop_motor,
		      struct ggiGA_resource_props *prop_carb,
		      struct ggiGA_resource_props *prop_tank,
		      enum ggiGA_resource_type type)
{
	ggiGA_resource_handle reqhandle_motor = NULL;
	ggiGA_resource_handle reqhandle_tank = NULL;
	ggiGA_resource_handle reqhandle_carb = NULL;

	ggiGA_resource_handle reshandle_motor = NULL;
	ggiGA_resource_handle reshandle_tank = NULL;
	ggiGA_resource_handle reshandle_carb = NULL;

	ggiGA_resource_handle reshandle2_motor = NULL;
	ggiGA_resource_handle reshandle2_tank = NULL;
	ggiGA_resource_handle reshandle2_carb = NULL;

	ggiGA_resource_list reslist = NULL;
	ggiGA_resource_list reslist2 = NULL;

	ggiOvl_t ovl;
	int rc;


	/* Prime the pump -- in case no mode resource in visual yet. */
	if (*(LIBGGI_OVLEXT(vis)->reqlist) == NULL) {
		ggiGAGet(vis, LIBGGI_OVLEXT(vis)->reqlist);
	}	/* if */

	/* Copy the current resource list into the local reslist1,
	 * and add our sprite to it, save the handle.
	 */
	rc = ggiGAGet(vis, &reslist);
	LIB_ASSERT(rc == GALLOC_OK);
	if (rc != GALLOC_OK) return NULL;

	rc = ggiOvlAdd(&reslist, prop_motor, prop_carb,
		       prop_tank, NULL, type,
		       &reshandle_motor, &reshandle_carb,
		       &reshandle_tank, NULL);
	LIB_ASSERT(rc == GALLOC_OK);
	if (rc != GALLOC_OK) goto err0;


	rc = ggiGACheck(vis, reslist, NULL);
	if (rc != GALLOC_OK) goto err0;


	rc = ggiGASet(vis, reslist, &reslist2);
	if (rc != GALLOC_OK) goto err1;


	/* Get the handle of the filled-out resource. */
	reshandle2_motor = ggiGAHandle(reslist2, reslist, reshandle_motor);
	LIB_ASSERT(reshandle2_tank != NULL);
	if (reshandle2_tank == NULL) goto err1;

	reshandle2_tank = ggiGAHandle(reslist2, reslist, reshandle_tank);
	LIB_ASSERT(reshandle2_tank != NULL);
	if (reshandle2_tank == NULL) goto err2;

	reshandle2_carb = ggiGAHandle(reslist2, reslist, reshandle_carb);
	LIB_ASSERT(reshandle2_tank != NULL);
	if (reshandle2_tank == NULL) goto err3;


	ovl = ggiOvlTie(vis, reshandle2_motor, reshandle2_carb, 
			reshandle2_tank);
	if (ovl == NULL) goto err4;


	rc = ggiOvlAdd(LIBGGI_OVLEXT(vis)->reqlist,
		       prop_motor, prop_carb, prop_tank, NULL, type,
		       &reqhandle_motor, &reqhandle_carb,
		       &reqhandle_tank, NULL);
	LIB_ASSERT(rc == GALLOC_OK);
	if (rc != GALLOC_OK) goto err5;

	ovl->req_handle_motor = reqhandle_motor;
	ovl->req_handle_tank = reqhandle_tank;
	ovl->req_handle_carb = reqhandle_carb;
	rc = ggiGAEmptyList(&reslist2);
	LIB_ASSERT(rc == GALLOC_OK);


	return ovl;

 err5:
	rc = ggiOvlUntie(&ovl);
	LIB_ASSERT(rc == GALLOC_OK);
 err4:
	rc = ggiGARelease(vis, &reslist2, &reshandle2_carb);
	LIB_ASSERT(rc == GALLOC_OK);
 err3:
	rc = ggiGARelease(vis, &reslist2, &reshandle2_tank);
	LIB_ASSERT(rc == GALLOC_OK);
 err2:
	rc = ggiGARelease(vis, &reslist2, &reshandle2_motor);
	LIB_ASSERT(rc == GALLOC_OK);
 err1:
	rc = ggiGAEmptyList(&reslist2);
	LIB_ASSERT(rc == GALLOC_OK);
 err0:
	rc = ggiGAEmptyList(&reslist);
	LIB_ASSERT(rc == GALLOC_OK);

	return NULL;
}	/* ggiOvlCreate */



ggiOvl_t ggiOvlCreateSprite(ggi_visual_t vis,
			    long width, long height,
			    enum ggiGA_resource_type subtype,
			    enum ggiOvl_flags flags)
{
	struct ggiGA_resource_props props_motor;
	struct ggiGA_resource_props props_tank;
	struct ggiGA_resource_props props_carb;


	ggiOvlClearMotorProperties(&props_motor);
	ggiOvlClearTankProperties(&props_tank);
	ggiOvlClearCarbProperties(&props_carb);


	props_tank.flags = flags;
	props_tank.size.area.x = width;
	props_tank.size.area.y = height;

	return ggiOvlCreate(vis, &props_motor, &props_carb,
			    &props_tank, subtype);
}	/* ggiOvlCreateSprite */


ggiOvl_t ggiOvlCreateWindow(ggi_visual_t vis,
			    long width, long height,
			    enum ggiGA_resource_type subtype,
			    enum ggiOvl_flags flags)
{
	struct ggiGA_resource_props props_motor;
	struct ggiGA_resource_props props_carb;
	struct ggiGA_resource_props props_tank;


	ggiOvlClearMotorProperties(&props_motor);
	ggiOvlClearCarbProperties(&props_carb);
	ggiOvlClearTankProperties(&props_tank);

	props_tank.flags = flags;
	props_tank.size.area.x = width;
	props_tank.size.area.y = height;

	return ggiOvlCreate(vis, &props_motor, &props_carb,
			    &props_tank, subtype);
}	/* ggiOvlCreateWindow */


ggiOvl_t ggiOvlCreateVideo(ggi_visual_t vis,
			   long width, long height,
			   enum ggiGA_resource_type subtype,
			   enum ggiOvl_flags flags)
{
	struct ggiGA_resource_props props_motor;
	struct ggiGA_resource_props props_carb;
	struct ggiGA_resource_props props_tank;


	ggiOvlClearMotorProperties(&props_motor);
	ggiOvlClearCarbProperties(&props_carb);
	ggiOvlClearTankProperties(&props_tank);

	props_tank.flags = flags;
	props_tank.size.area.x = width;
	props_tank.size.area.y = height;

        return ggiOvlCreate(vis, &props_motor, &props_carb,
	                    &props_tank, subtype);
}	/* ggiOvlCreateVideo */


int ggiOvlDestroyOvl(ggiOvl_t *ovl)
{
	ggiGA_resource_handle handle_tank = NULL;
	ggiGA_resource_handle handle_carb = NULL;
	ggiGA_resource_handle handle_motor = NULL;
	ggi_visual_t vis;
	int rc;


	if (ovl == NULL) return GGI_EARGINVAL;
	if (ovl[0] == NULL) return GGI_EARGINVAL;


	LIB_ASSERT(ovl[0]->vis == ggiGAGetVisual(ovl[0]->res_handle_tank));
	vis = ovl[0]->vis;

	handle_tank = ovl[0]->req_handle_tank;
	LIB_ASSERT(handle_tank != NULL);
	handle_carb = ovl[0]->req_handle_carb;
	LIB_ASSERT(handle_carb != NULL);
	handle_motor = ovl[0]->req_handle_motor;
	LIB_ASSERT(handle_motor != NULL);

	rc = ggiOvlUntie(ovl);
	LIB_ASSERT(rc == GALLOC_OK);

	rc = ggiGARelease(vis, LIBGGI_OVLEXT(vis)->reqlist, &handle_tank);
	LIB_ASSERT(rc == GALLOC_OK);
	rc = ggiGARelease(vis, LIBGGI_OVLEXT(vis)->reqlist, &handle_carb);
	LIB_ASSERT(rc == GALLOC_OK);

	if (ggiGAMotorGetTank(LIBGGI_OVLEXT(vis)->reqlist[0], handle_motor, NULL) == NULL) {
		rc = ggiGARelease(vis, LIBGGI_OVLEXT(vis)->reqlist,
				  &handle_motor);
		LIB_ASSERT(rc == GALLOC_OK);
	}	/* if */

	return rc;
}	/* ggiOvlDestroyOvl */
